"""Attribute, Dimension, and Schema
================================

Classes for accessing SciDB data and schemas.

"""

import itertools
import numpy
import pandas
import re
import six
import struct
import warnings


type_map_numpy = dict(
    (k, numpy.dtype(v)) for (k, v) in
    [(t.__name__, t) for t in (
        numpy.bool,

        numpy.int8,
        numpy.int16,
        numpy.int32,
        numpy.int64,

        numpy.uint8,
        numpy.uint16,
        numpy.uint32,
        numpy.uint64,
    )] + [
        ('char', 'S1'),
        ('double', numpy.float64),
        ('float', numpy.float32),
        ('string', numpy.object),
        ('binary', numpy.object),
        ('datetime', 'datetime64[s]'),
        ('datetimetz', [('time', 'datetime64[s]'),
                        ('tz', 'timedelta64[s]')]),
    ])

type_map_inv_numpy = {v: k
                      for k, v in six.iteritems(type_map_numpy)
                      if v != numpy.dtype(numpy.object)}
type_map_inv_numpy.update(dict(
    (numpy.dtype(k), v) for (k, v) in
    [
        (numpy.str_, 'string'),
        (numpy.string_, 'string'),
        (numpy.datetime64, 'datetime'),
        (numpy.timedelta64, 'datetimetz'),
    ]))

type_map_struct = {
    'bool': '?',

    'char': 'c',

    'int8': 'b',
    'int16': '<h',
    'int32': '<i',
    'int64': '<q',

    'float': '<f',
    'double': '<d',

    'datetime': '<q',
    'datetimetz': '<qq',
    }

# Add uint types
for key in list(type_map_struct.keys()):
    if key.startswith('int'):
        type_map_struct['u' + key] = type_map_struct[key].upper()

# Add null-able type
for (key, val) in type_map_struct.items():
    if len(val) > 1:
        val_null = val[0] + 'B' + val[1]
    else:
        val_null = 'B' + val
    type_map_struct[key] = (val, val_null)

# Type promotion map for Pandas DataFrame
# http://pandas.pydata.org/pandas-docs/stable/gotchas.html#na-type-promotions
type_map_promo = dict(
    (k, numpy.dtype(v)) for (k, v) in
    [
        ('bool', numpy.object),
        ('char', numpy.object),

        ('int8', numpy.float16),
        ('int16', numpy.float32),
        ('int32', numpy.float64),
        ('int64', numpy.float64),

        ('uint8', numpy.float16),
        ('uint16', numpy.float32),
        ('uint32', numpy.float64),
        ('uint64', numpy.float64),
    ])

one_att_name = 'x'
one_dim_name = 'i'


class Attribute(object):
    """Represent SciDB array attribute

    Construct an attribute using Attribute constructor:

    >>> Attribute('foo', 'int64', not_null=True)
    ... # doctest: +NORMALIZE_WHITESPACE
    Attribute(name='foo',
              type_name='int64',
              not_null=True,
              default=None,
              compression=None)

    >>> Attribute('foo', 'int64', default=100, compression='zlib')
    ... # doctest: +NORMALIZE_WHITESPACE
    Attribute(name='foo',
              type_name='int64',
              not_null=False,
              default=100,
              compression='zlib')


    Construct an attribute from a string:

    >>> Attribute.fromstring('foo:int64')
    ... # doctest: +NORMALIZE_WHITESPACE
    Attribute(name='foo',
              type_name='int64',
              not_null=False,
              default=None,
              compression=None)

    >>> Attribute.fromstring(
    ...     "taz : string NOT null DEFAULT '' compression 'bzlib'")
    ... # doctest: +NORMALIZE_WHITESPACE
    Attribute(name='taz',
              type_name='string',
              not_null=True,
              default="''",
              compression='bzlib')
    """

    _regex = re.compile('''
        \\s*
        (?P<name>      \\w+ )  \\s* : \\s*
        (?P<type_name> \\w+ )  \\s*
        (?:                   (?P<not_null>    NOT )? \\s+ NULL )? \\s*
        (?: DEFAULT     \\s+  (?P<default>     \\S+ )           )? \\s*
        (?: COMPRESSION \\s+ '(?P<compression> \\w+ )'          )? \\s*
        $''', re.VERBOSE | re.IGNORECASE)
    # length dtype for variable-size SciDB types
    _length_dtype = numpy.dtype(numpy.uint32)
    _length_fmt = '<I'

    def __init__(self,
                 name,
                 type_name,
                 not_null=False,
                 default=None,
                 compression=None):
        self.__name = name
        self.type_name = type_name
        self.not_null = bool(not_null)
        self.default = default
        self.compression = compression

        self.fmt_scidb = '{}{}'.format(self.type_name,
                                       '' if self.not_null else ' null')
        self.fmt_struct = type_map_struct.get(self.type_name, None)

        self._set_dtype()

    def __iter__(self):
        return (i for i in (
            self.name,
            self.type_name,
            self.not_null,
            self.default,
            self.compression))

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __repr__(self):
        return ('{}(' +
                'name={!r}, ' +
                'type_name={!r}, ' +
                'not_null={!r}, ' +
                'default={!r}, ' +
                'compression={!r})').format(
            type(self).__name__, *self)

    def __str__(self):
        return '{}:{}{}{}{}'.format(
            self.name,
            self.type_name,
            ' NOT NULL' if self.not_null else '',
            ' DEFAULT {}'.format(self.default) if self.default else '',
            " COMPRESSION '{}'".format(self.compression)
            if self.compression else '')

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value
        self._set_dtype()

    def _set_dtype(self):
        self.dtype_val = type_map_numpy.get(self.type_name, numpy.object)
        # >>> numpy.dtype([(u"a", int)])
        # TypeError: data type not understood
        # https://github.com/numpy/numpy/issues/2407
        # cannot use `self.name` directly, use `str(...)`
        if self.not_null:
            self.dtype = numpy.dtype([(str(self.name), self.dtype_val)])
        else:
            self.dtype = numpy.dtype([(str(self.name),
                                       [('null', numpy.uint8),
                                        ('val', self.dtype_val)])])

    def is_fixsize(self):
        return self.dtype_val != numpy.object

    def itemsize(self, buf=None, offset=0):
        if self.dtype_val != numpy.object:
            return self.dtype.itemsize

        null_size = 0 if self.not_null else 1
        value_size = numpy.frombuffer(
            buf, numpy.uint32, 1, offset + null_size)[0]
        return null_size + Attribute._length_dtype.itemsize + value_size

    def frombytes(self, buf, offset=0, size=None, promo=False):
        null_size = 0 if self.not_null else 1

        if self.dtype_val == numpy.object:
            if self.type_name == 'string':
                val = buf[offset + null_size +
                          Attribute._length_dtype.itemsize:
                          offset + size - 1].decode('utf-8')
            else:
                val = buf[offset + null_size +
                          Attribute._length_dtype.itemsize:
                          offset + size]
        else:
            val = struct.unpack(
                self.fmt_struct[0], buf[offset + null_size:offset + size])
            if len(val) == 1:
                val = val[0]

        if self.not_null:
            return val
        else:
            missing = struct.unpack('B', buf[offset:offset + null_size])[0]
            if promo:
                return val if missing == 255 else None
            else:
                return (missing, val)

    def tobytes(self, val):
        if self.dtype_val == numpy.object:
            if self.type_name == 'string':
                buf = b''.join(
                    [struct.pack(Attribute._length_fmt, len(val) + 1),
                     val.encode(),
                     b'\x00'])
            elif self.type_name == 'binary':
                buf = b''.join(
                    [struct.pack(Attribute._length_fmt, len(val)), val])
            else:
                raise NotImplementedError('Convert <{}> to bytes'.format(self))
        else:
            if self.not_null:
                buf = struct.pack(self.fmt_struct[0], val)
            else:
                if isinstance(val, numpy.void):
                    # NumPy structured array
                    buf = struct.pack(self.fmt_struct[1], *val)
                else:
                    buf = struct.pack(self.fmt_struct[1], 255, val)
        return buf

    @classmethod
    def fromstring(cls, string):
        try:
            return cls(**Attribute._regex.match(string).groupdict())
        except AttributeError:
            raise Exception('Failed to parse attribute: {}'.format(string))

    @classmethod
    def fromdtype(cls, dtype_descr):
        if isinstance(dtype_descr[1], str):
            # e.g. ('name', 'int64')
            dtype_val = dtype_descr[1]
            not_null = True
        else:
            # e.g. ('name', [('null': 'int8'), ('val': 'int64')]
            #      ('name', [('time', 'datetime64'), ('tz', 'timedelta64')])
            #      ('name', [('null': 'int8'),
            #                ('val' : [('time', 'datetime64'),
            #                          ('tz', 'timedelta64')])])
            if dtype_descr[1][0][0] == 'null':
                not_null = False
                dtype_val = dtype_descr[1][1][1]
            else:
                not_null = True
                dtype_val = dtype_descr[1]

        dtype_val = numpy.dtype(dtype_val)
        if dtype_val in type_map_inv_numpy.keys():
            type_name = type_map_inv_numpy[dtype_val]
        else:
            # if dtype_val not found in map, try the dtype_val.type
            # (without the length)
            ty = numpy.dtype(dtype_val.type)
            # e.g. '<U3' --type--> '<U' --map--> numpy.str_
            if ty in type_map_inv_numpy.keys():
                type_name = type_map_inv_numpy[ty]
            else:
                raise Exception(
                    'No SciDB type mapping for NumPy type {}'.format(
                        dtype_val))

        return cls(name=dtype_descr[0] if dtype_descr[0] else one_att_name,
                   type_name=type_name,
                   not_null=not_null)


class Dimension(object):
    """Represent SciDB array dimension

    Construct a dimension using the Dimension constructor:

    >>> Dimension('foo')
    ... # doctest: +NORMALIZE_WHITESPACE
    Dimension(name='foo',
              low_value=None,
              high_value=None,
              chunk_overlap=None,
              chunk_length=None)

    >>> Dimension('foo', -100, '10', '?', '1000')
    ... # doctest: +NORMALIZE_WHITESPACE
    Dimension(name='foo',
              low_value=-100,
              high_value=10,
              chunk_overlap='?',
              chunk_length=1000)


    Construct a dimension from a string:

    >>> Dimension.fromstring('foo')
    ... # doctest: +NORMALIZE_WHITESPACE
    Dimension(name='foo',
              low_value=None,
              high_value=None,
              chunk_overlap=None,
              chunk_length=None)

    >>> Dimension.fromstring('foo=-100:*:?:10')
    ... # doctest: +NORMALIZE_WHITESPACE
    Dimension(name='foo',
              low_value=-100,
              high_value='*',
              chunk_overlap='?',
              chunk_length=10)
    """

    _regex = re.compile('''
        \\s*
        (?P<name> \\w+ ) \\s*
        (?: = \\s* (?P<low_value>  [^:\\s]+ ) \\s*
                   : \\s*
                   (?P<high_value> [^:\\s]+ ) \\s*
                   (?: : \\s* (?P<chunk_overlap> [^:\\s]+ ) \\s*
                              (?: : \\s* (?P<chunk_length> [^:\\s]+ ) )?
                   )?
        )?
        \\s* $''', re.VERBOSE)

    def __init__(self,
                 name,
                 low_value=None,
                 high_value=None,
                 chunk_overlap=None,
                 chunk_length=None):
        self.name = name

        try:
            self.low_value = int(low_value)
        except (TypeError, ValueError):
            self.low_value = low_value

        try:
            self.high_value = int(high_value)
        except (TypeError, ValueError):
            self.high_value = high_value

        try:
            self.chunk_overlap = int(chunk_overlap)
        except (TypeError, ValueError):
            self.chunk_overlap = chunk_overlap

        try:
            self.chunk_length = int(chunk_length)
        except (TypeError, ValueError):
            self.chunk_length = chunk_length

    def __iter__(self):
        return (i for i in (
            self.name,
            self.low_value,
            self.high_value,
            self.chunk_overlap,
            self.chunk_length))

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __repr__(self):
        return ('{}(' +
                'name={!r}, ' +
                'low_value={!r}, ' +
                'high_value={!r}, ' +
                'chunk_overlap={!r}, ' +
                'chunk_length={!r})').format(
            type(self).__name__, *self)

    def __str__(self):
        out = self.name
        if self.low_value is not None:
            out += '={}:{}'.format(self.low_value, self.high_value)
            if self.chunk_overlap is not None:
                out += ':{}'.format(self.chunk_overlap)
                if self.chunk_length is not None:
                    out += ':{}'.format(self.chunk_length)
        return out

    @classmethod
    def fromstring(cls, string):
        try:
            return cls(**Dimension._regex.match(string).groupdict())
        except AttributeError:
            raise Exception('Failed to parse dimension: {}'.format(string))


class Schema(object):
    """Represent SciDB array schema

    Construct a schema using Schema, Attribute, and Dimension
    constructors:

    >>> Schema('foo', (Attribute('x', 'int64'),), (Dimension('i', 0, 10),))
    ... # doctest: +NORMALIZE_WHITESPACE
    Schema(name='foo',
           atts=(Attribute(name='x',
                           type_name='int64',
                           not_null=False,
                           default=None,
                           compression=None),),
           dims=(Dimension(name='i',
                           low_value=0,
                           high_value=10,
                           chunk_overlap=None,
                           chunk_length=None),))


    Construct a schema using Schema constructor and fromstring methods
    of Attribute and Dimension:

    >>> Schema('foo',
    ...        (Attribute.fromstring('x:int64'),),
    ...        (Dimension.fromstring('i=0:10'),))
    ... # doctest: +NORMALIZE_WHITESPACE
    Schema(name='foo',
           atts=(Attribute(name='x',
                           type_name='int64',
                           not_null=False,
                           default=None,
                           compression=None),),
           dims=(Dimension(name='i',
                           low_value=0,
                           high_value=10,
                           chunk_overlap=None,
                           chunk_length=None),))


    Construct a schema from a string:

    >>> Schema.fromstring(
    ...     'foo@1<x:int64 not null, y:double>[i=0:*; j=-100:0:0:10]')
    ... # doctest: +NORMALIZE_WHITESPACE
    Schema(name='foo@1',
           atts=(Attribute(name='x',
                           type_name='int64',
                           not_null=True,
                           default=None,
                           compression=None),
                 Attribute(name='y',
                           type_name='double',
                           not_null=False,
                           default=None,
                           compression=None)),
           dims=(Dimension(name='i',
                           low_value=0,
                           high_value='*',
                           chunk_overlap=None,
                           chunk_length=None),
                 Dimension(name='j',
                           low_value=-100,
                           high_value=0,
                           chunk_overlap=0,
                           chunk_length=10)))


    Print a schema constructed from a string:

    >>> print(Schema.fromstring('<x:int64,y:float> [i=0:2:0:1000000; j=0:*]'))
    ... # doctest: +NORMALIZE_WHITESPACE
    <x:int64,y:float> [i=0:2:0:1000000; j=0:*]


    Format Schema object to only print the schema part without the
    array name:

    >>> '{:h}'.format(Schema.fromstring('foo<x:int64>[i]'))
    '<x:int64> [i]'
    """

    _regex_name = re.compile(
        '\\s* (?: not \\s+ empty \\s+ )? (?P<name> [\\w@]+ )?', re.VERBOSE)

    _regex_atts = re.compile(
        '\\s*  < ( [^,>]+  \\s* (?: , \\s* [^,>]+  \\s* )* )  >', re.VERBOSE)

    _regex_dims = re.compile(
        '\\s* \\[ ( [^;\\]]+ \\s* (?: ; \\s* [^;\\]]+ \\s* )* ) \\] \\s*',
        re.VERBOSE)

    def __init__(self, name=None, atts=(), dims=()):
        self.name = name
        self.atts = tuple(atts)
        self.dims = tuple(dims)

        # Set lazy
        self.__atts_dtype = None
        self.__atts_fmt_scidb = None

    def __iter__(self):
        return (i for i in (self.name, ) + self.atts + self.dims)

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __repr__(self):
        return '{}(name={!r}, atts={!r}, dims={!r})'.format(
            type(self).__name__, self.name, self.atts, self.dims)

    def __str__(self):
        return self._render()

    def __format__(self, fmt_spec=''):
        return self._render(no_name='h' in fmt_spec)

    def _render(self, no_name=False):
        return '{}<{}> [{}]'.format(
            self.name if not no_name and self.name else '',
            ','.join(str(a) for a in self.atts),
            '; '.join(str(d) for d in self.dims))

    def _promo_warning(self):
        cnt = sum(not a.not_null for a in self.atts)
        if cnt:
            warnings.warn(
                ('{} type(s) promoted for null support.' +
                 ' Precision loss may occur').format(cnt),
                stacklevel=2)

    @property
    def atts_dtype(self):
        if self.__atts_dtype is None:
            self.__atts_dtype = numpy.dtype(list(itertools.chain.from_iterable(
                a.dtype.descr for a in self.atts)))
        return self.__atts_dtype

    @property
    def atts_fmt_scidb(self):
        if self.__atts_fmt_scidb is None:
            self.__atts_fmt_scidb = '({})'.format(
                ', '.join(a.fmt_scidb for a in self.atts))
        return self.__atts_fmt_scidb

    def pprint(self):
        print(self)
        info = numpy.empty(
            (len(self.atts) + len(self.dims),),
            dtype=[('name', numpy.object),
                   ('class', numpy.object),
                   ('type', numpy.object),
                   ('nullable', numpy.object),
                   ('start', numpy.object),
                   ('end', numpy.object),
                   ('overlap', numpy.object),
                   ('chunk', numpy.object)])
        pos = 0
        for a in self.atts:
            info.put((pos,),
                     (a.name,
                      'attr',
                      a.type_name,
                      not a.not_null,
                      '',
                      '',
                      '',
                      ''))
            pos += 1
        for d in self.dims:
            info.put((pos,),
                     (d.name,
                      'dim',
                      'int64',
                      '',
                      d.low_value,
                      d.high_value,
                      d.chunk_overlap,
                      d.chunk_length))
            pos += 1
        print(pandas.DataFrame.from_records(info))

    def is_fixsize(self):
        return all(a.is_fixsize() for a in self.atts)

    def make_unique(self):
        """Make dimension and attribute names unique within the schema. Return
        ``True`` if any dimension or attribute was renamed.

        >>> s = Schema(None, (Attribute('i', 'bool'),), (Dimension('i'),))
        >>> print(s)
        <i:bool> [i]
        >>> s.make_unique()
        True
        >>> print(s)
        <i:bool> [i_1]

        >>> s = Schema.fromstring('<i:bool, i:int64>[i;i_1;i]')
        >>> s.make_unique()
        True
        >>> print(s)
        <i:bool,i_2:int64> [i_3; i_1; i_4]

        """
        all_before = set(itertools.chain((a.name for a in self.atts),
                                         (d.name for d in self.dims)))

        # Check if overall duplicates are present
        if len(all_before) < len(self.atts) + len(self.dims):

            all_after = set()

            # Process attributes
            for a in self.atts:
                # Start renaming after the first copy. First copy
                # will not be in all_after. From second copy
                # on-wards, a copy will be in all_after.
                if a.name in all_after:
                    new_name_tmpl = a.name + '_{}'
                    count = 1
                    new_name = new_name_tmpl.format(count)
                    while (new_name in all_before or
                           new_name in all_after):
                        count += 1
                        new_name = new_name_tmpl.format(count)
                    a.name = new_name
                all_after.add(a.name)

            # Process dimensions
            for d in self.dims:
                if d.name in all_after:
                    new_name_tmpl = d.name + '_{}'
                    count = 1
                    new_name = new_name_tmpl.format(count)
                    while (new_name in all_before or
                           new_name in all_after):
                        count += 1
                        new_name = new_name_tmpl.format(count)
                    d.name = new_name
                all_after.add(d.name)

            # Reset dtype
            self.__atts_dtype = None

            return True
        else:
            return False

    def make_dims_atts(self):
        """Make attributes from dimensions and pre-append them to the
        attributes list.

        >>> s = Schema(None, (Attribute('x', 'bool'),), (Dimension('i'),))
        >>> print(s)
        <x:bool> [i]
        >>> s.make_dims_atts()
        >>> print(s)
        <i:int64 NOT NULL,x:bool> [i]

        >>> s = Schema.fromstring('<x:bool>[i;j]')
        >>> s.make_dims_atts()
        >>> print(s)
        <i:int64 NOT NULL,j:int64 NOT NULL,x:bool> [i; j]

        """
        self.atts = tuple(itertools.chain(
            (Attribute(d.name, 'int64', not_null=True) for d in self.dims),
            self.atts))

        # Reset
        self.__atts_dtype = None
        self.__atts_fmt_scidb = None

    def get_promo_atts_dtype(self):
        self._promo_warning()
        return numpy.dtype(
            [a.dtype.descr[0] if a.not_null else
             (a.dtype.names[0],
              type_map_promo.get(
                  a.type_name, type_map_numpy.get(a.type_name, numpy.object)))
             for a in self.atts])

    def promote(self, data):
        """Promote nullable attributes in the DataFrame to types which
        support some type of null values as per Pandas `promotion
        scheme <http://pandas.pydata.org/pandas-docs/stable/gotchas.html
        #na-type-promotions>`_

        """
        self._promo_warning()
        for a in self.atts:
            if not a.not_null:
                data[a.name] = pandas.Series(
                    data=[attr[1] if attr[0] == 255 else numpy.NAN
                          for attr in data[a.name]],
                    dtype=type_map_promo.get(
                        a.type_name, type_map_numpy.get(
                            a.type_name, numpy.object)))

    def frombytes(self, buf, as_dataframe=False, dataframe_promo=True):
        # Scan content and build (offset, size) metadata
        off = 0
        buf_meta = []
        while off < len(buf):
            meta = []
            for att in self.atts:
                sz = att.itemsize(buf, off)
                meta.append((off, sz))
                off += sz
            buf_meta.append(meta)

        # Create NumPy record array
        if as_dataframe and dataframe_promo:
            data = numpy.empty((len(buf_meta),),
                               dtype=self.get_promo_atts_dtype())
        else:
            data = numpy.empty((len(buf_meta),), dtype=self.atts_dtype)

        # Extract values using (offset, size) metadata
        # Populate NumPy record array
        pos = 0
        for meta in buf_meta:
            data.put((pos,),
                     tuple(att.frombytes(
                         buf,
                         off,
                         sz,
                         promo=as_dataframe and dataframe_promo)
                           for (att, (off, sz)) in zip(self.atts, meta)))
            pos += 1
        return data

    def tobytes(self, data):
        buf_lst = []
        if len(data.dtype) > 0:
            # NumPy structured array
            if len(self.atts_dtype) == 1:
                # One attribute
                atr = self.atts[0]
                for cell in data:
                    buf_lst.append(atr.tobytes(cell[0]))
            else:
                # Multiple attributes
                for cell in data:
                    for (atr, val) in zip(self.atts, cell):
                        buf_lst.append(atr.tobytes(val))
        else:
            # NumPy single-field array
            atr = self.atts[0]
            for val in data:
                buf_lst.append(atr.tobytes(val))
        return b''.join(buf_lst)

    @classmethod
    def fromstring(cls, string):
        name_match = Schema._regex_name.match(string)
        atts_match = Schema._regex_atts.match(string, name_match.end(0))
        dims_match = Schema._regex_dims.match(string, atts_match.end(0))

        name = name_match.groupdict()['name']
        return cls(
            name.strip() if name else None,
            (Attribute.fromstring(s)
             for s in atts_match.group(1).split(',')),
            (Dimension.fromstring(s)
             for s in dims_match.group(1).split(';')))

    @classmethod
    def fromdtype(cls, dtype):
        return cls(
            None,
            (Attribute.fromdtype(dt) for dt in dtype.descr),
            (Dimension(one_dim_name),))


if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.REPORT_ONLY_FIRST_FAILURE)



"""Quick Start
===========

Install the package with::

  $ pip install scidb-py

Connect to SciDB and run a query:

>>> from scidbpy import connect
>>> db = connect('http://localhost:8080')
>>> db.iquery('store(build(<x:int64>[i=0:2], i), foo)')

Download data from SciDB:

>>> db.arrays.foo[:]
   i    x
0  0  0.0
1  1  1.0
2  2  2.0

Upload data to SciDB and create an array:

>>> import numpy
>>> ar = db.input(upload_data=numpy.arange(3)).store()
>>> print(ar)
... # doctest: +ELLIPSIS
py_..._1

Run a query with chained operators and download the resulting array:

>>> db.join(ar, 'foo').apply('j', ar.i + 1)[:]
   i  x  x_1  j
0  0  0  0.0  1
1  1  1  1.0  2
2  2  2  2.0  3

Operators can also be composed:

>>> db.apply(db.join(ar, 'foo'), 'j', ar.i + 1)[:]
   i  x  x_1  j
0  0  0  0.0  1
1  1  1  1.0  2
2  2  2  2.0  3

Cleanup:

>>> db.remove(db.arrays.foo)
>>> del ar


Requirements
============

SciDB ``16.9`` or newer with Shim

Python ``2.7.x``, ``3.4.x``, ``3.5.x``, ``3.6.x`` or newer.

Required Python packages::

  backports.weakref
  enum34
  numpy
  pandas
  requests
  six


Installation
============

Install latest release::

  $ pip install scidb-py

Install development version from GitHub::

  $ pip install git+http://github.com/paradigm4/scidb-py.git


Connect to SciDB
================

Connect to SciDB using :func:`connect()<scidbpy.db.connect>`:

>>> from scidbpy import connect
>>> db = connect()
>>> db = connect('http://localhost:8080')

``connect()`` is aliased to the constructor of the
:class:`DB<scidbpy.db.DB>` class. See :meth:`DB()<scidbpy.db.DB>` for
the complete set of arguments that can be provided to
``connect``. ``http://localhost:8080`` is the default connection URL
is none is provided.

Display information about the ``db`` object:

>>> db
DB('http://localhost:8080', None, None, None, False, None, False, 256, False)

>>> print(db)
scidb_url         = http://localhost:8080
scidb_auth        = None
http_auth         = None
verify            = None
admin             = False
namespace         = None
use_arrow         = False
result_size_limit = 256
no_ops            = False


Advanced Connection
-------------------

Provide `Shim <https://github.com/Paradigm4/shim>`_ credentials:

>>> db = connect(http_auth=('foo', 'bar'))

>>> db
... # doctest: +NORMALIZE_WHITESPACE
DB('http://localhost:8080',
   None,
   ('foo', PASSWORD_PROVIDED),
   None,
   False,
   None,
   False,
   256,
   False)

>>> print(db)
scidb_url         = http://localhost:8080
scidb_auth        = None
http_auth         = ('foo', PASSWORD_PROVIDED)
verify            = None
admin             = False
namespace         = None
use_arrow         = False
result_size_limit = 256
no_ops            = False

To prompt the user for the password, use:

>>> import getpass
>>> db = connect(http_auth=('foo', getpass.getpass()))
... # doctest: +SKIP
Password:


Use SSL:

>>> db_ssl = connect('https://localhost:8083', verify=False)

>>> print(db_ssl)
scidb_url         = https://localhost:8083
scidb_auth        = None
http_auth         = None
verify            = False
admin             = False
namespace         = None
use_arrow         = False
result_size_limit = 256
no_ops            = False

See Python `requests <http://docs.python-requests.org/en/master/>`_
library `SSL Cert Verification
<http://docs.python-requests.org/en/master/user/advanced/
#ssl-cert-verification>`_ section for details on the ``verify``
argument. ``verify=False`` disables SSL certificate
verification. Warnings about the unverified HTTPS requests are
displayed. The warnings can be disabled as well by either setting
``PYTHONWARNINGS="ignore:Unverified HTTPS request"`` in the
environment before starting Python or by doing:

>>> import requests
>>> requests.packages.urllib3.disable_warnings(
...   requests.packages.urllib3.exceptions.InsecureRequestWarning)


Use SSL and SciDB credentials:

>>> db_ssl = connect(
...   'https://localhost:8083', scidb_auth=('foo', 'bar'), verify=False)

>>> print(db_ssl)
scidb_url         = https://localhost:8083
scidb_auth        = ('foo', PASSWORD_PROVIDED)
http_auth         = None
verify            = False
admin             = False
namespace         = None
use_arrow         = False
result_size_limit = 256
no_ops            = False


When using the ``iquery`` SciDB client, the ``--admin`` flag is
available for opening a higher-priority session. This flag is also
available in SciDB-Py and can be set at connection time (see `SciDB
Documentation <https://paradigm4.atlassian.net/wiki/spaces/scidb>`_
for details on the effects of the flag). By default this flag is set
to ``False``:

>>> db_admin = connect(admin=True)

>>> print(db_admin)
scidb_url         = http://localhost:8080
scidb_auth        = None
http_auth         = None
verify            = None
admin             = True
namespace         = None
use_arrow         = False
result_size_limit = 256
no_ops            = False


By default, the ``connect`` function queries SciDB for the list of
available operators. This list is used for easy access to the SciDB
operators, see the *SciDB Operators* section below. As a consequence
the connection to SciDB is verified and a small delay might
occur. This behavior can be disabled using the ``no_ops=True``
parameter. Accessing the SciDB operators as described in *SciDB
Operators* will not be possible until the ``load_ops()`` function is
called on the ``DB`` instance:

>>> db_no_ops = connect(no_ops=True)
>>> db_no_ops.scan
Traceback (most recent call last):
    ...
AttributeError: Operators not loaded. Run 'load_ops()' or use \
'no_ops=False' (default) at connection time (constructor)

No query has been issued to SciDB yet.

>>> db_no_ops.load_ops()
>>> db_no_ops.scan
... # doctest: +NORMALIZE_WHITESPACE
Operator(db=DB('http://localhost:8080',
               None,
               None,
               None,
               False,
               None,
               False,
               256,
               False),
         name='scan',
         args=[])


SciDB Arrays
============

SciDB arrays can be accessed using ``DB.arrays``:

>>> db = connect()
>>> db.iquery('store(build(<x:int64>[i=0:2], i), foo)')

>>> dir(db.arrays)
... # doctest: +ALLOW_UNICODE
[u'foo']

>>> dir(db.arrays.foo)
... # doctest: +ALLOW_UNICODE
[u'i', u'x']

>>> db.arrays.foo[:]
   i    x
0  0  0.0
1  1  1.0
2  2  2.0

>>> db.arrays['foo'][:]
   i    x
0  0  0.0
1  1  1.0
2  2  2.0

To get the schema of an array, we can use the ``schema`` utility
function:

>>> print(db.arrays.foo.schema())
foo<x:int64> [i=0:2:0:1000000]

>>> db.arrays.foo.schema().pprint()
... # doctest: +NORMALIZE_WHITESPACE
foo<x:int64> [i=0:2:0:1000000]
  name class   type nullable start end overlap    chunk
0    x  attr  int64     True
1    i   dim  int64              0   2       0  1000000

>>> db.iquery('remove(foo)')

>>> dir(db.arrays)
[]

Arrays specified explicitly are not checked:

>>> print(db.arrays.foo)
foo
>>> print(db.arrays.bar)
bar

In IPython, we can use <TAB> for auto-completion of array names,
array dimensions, and array attributes::

    In [1]: db.arrays.<TAB>
    In [1]: db.arrays.foo
    In [2]: db.arrays.foo.<TAB>
    In [2]: db.arrays.foo.x


SciDB Operators
===============

At connection time, the library downloads the list of available SciDB
operators and macros and makes them available through the ``DB`` class
instance:

>>> dir(db)
... # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS +ALLOW_UNICODE
[u'add_attributes',
 ...
 u'xgrid']

>>> db.apply
... # doctest: +NORMALIZE_WHITESPACE
Operator(db=DB('http://localhost:8080',
               None,
               None,
               None,
               False,
               None,
               False,
               256,
               False),
         name='apply',
         args=[])

>>> print(db.apply)
apply()

>>> db.missing
Traceback (most recent call last):
    ...
AttributeError: 'DB' object has no attribute 'missing'

In IPython, we can use <TAB> for auto-completion of operator names::

    In [1]: db.<TAB>
    In [1]: db.apply

The operators can be execute immediately or can be lazy and executed
at a later time. Operators that return arrays are lazy operators (e.g.,
``apply``, ``scan``, etc.). Operators which do not return arrays
execute immediately (e.g., ``create_array``, ``remove``, etc.).

>>> db.create_array('foo', '<x:int64>[i]')
>>> dir(db.arrays)
... # doctest: +ALLOW_UNICODE
[u'foo']

>>> db.remove(db.arrays.foo)
>>> dir(db.arrays)
[]

The list of available operators is re-loaded automatically when
a ``load_library`` query is issued:

>>> db.load_library('limit')
... # doctest: +SKIP
>>> 'limit' in dir(db)
... # doctest: +SKIP
True

A similar functionality is **not** implemented for ``unload_library``
operator. The ``unload_library`` operator requires a SciDB restart,
which makes re-loading the list of operators not
practical. Nevertheless, one can trigger the re-loading manually after
SciDB restart without creating a new ``DB`` instance:

>>> db.iquery("unload_library('limit')")
... # doctest: +SKIP

After SciDB restart:

>>> db.load_ops()
>>> 'limit' in dir(db)
... # doctest: +SKIP
False

The ``cross_join`` operator in SciDB supports aliasing for the array
arguments, e.g., ``cross_join(left_array as left_alias,...``. Aliasing
is possible in SciDB-Py using the ``%`` operator:

>>> db.iquery('store(build(<x:int64>[i=0:1], i), foo)')
>>> db.cross_join(db.arrays.foo % 'f1',
...               db.arrays.foo % 'f2',
...               'f1.i', 'f2.i')[:]
   i    x  x_1
0  0  0.0  0.0
1  1  1.0  1.0

>>> db.remove(db.arrays.foo)

To retrieve the schema of a query result, the ``schema`` utility
function can be used:

>>> print(db.build('<x:int64>[i=0:2]', 'i').schema())
build<x:int64> [i=0:2:0:1000000]

>>> db.build('<x:int64>[i=0:2]', 'i').schema().pprint()
... # doctest: +NORMALIZE_WHITESPACE
build<x:int64> [i=0:2:0:1000000]
  name class   type nullable start end overlap    chunk
0    x  attr  int64     True
1    i   dim  int64              0   2       0  1000000



Download Data from SciDB
------------------------

>>> db.build('<x:int8 not null>[i=0:2]', 'i + 10')[:]
   i   x
0  0  10
1  1  11
2  2  12

>>> db.build('<x:int8 not null>[i=0:2]', 'i + 10').fetch(as_dataframe=False)
... # doctest: +NORMALIZE_WHITESPACE
array([(0, 10), (1, 11), (2, 12)],
      dtype=[('i', '<i8'), ('x', 'i1')])

>>> db.build('<x:int8 not null>[i=0:2]', 'i + 10').fetch(
...     atts_only=True)
    x
0  10
1  11
2  12

>>> db.build('<x:int8 not null>[i=0:2]', 'i + 10').apply('y', 'x - 5')[:]
   i   x  y
0  0  10  5
1  1  11  6
2  2  12  7

>>> db.build('<x:int8 not null>[i=0:2]', 'i + 10').store('foo')
... # doctest: +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'foo')

>>> db.scan(db.arrays.foo)[:]
   i   x
0  0  10
1  1  11
2  2  12

>>> db.apply(db.arrays.foo, 'y', db.arrays.foo.x + 1)[:]
   i   x   y
0  0  10  11
1  1  11  12
2  2  12  13

>>> db.remove(db.arrays.foo)


Upload Data to SciDB
--------------------

``input`` and ``load`` operators can be used to upload data. An upload
schema can also be provided. If the resulting array or schema is not
provided, it can be generated from the upload data or upload
schema. If the upload format is not provided, it can be constructed
from the upload schema, upload data, or resulting array schema.

>>> db.input('<x:int64>[i]', upload_data=numpy.arange(3))[:]
   i    x
0  0  0.0
1  1  1.0
2  2  2.0

>>> db.input('<x:int64>[i]', upload_data=numpy.arange(3)).store(db.arrays.foo)
... # doctest: +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'foo')

>>> db.load(db.arrays.foo, upload_data=numpy.arange(3))
... # doctest: +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'foo')

>>> db.input('<x:int64>[j]', upload_data=numpy.arange(3, 6)
...  ).apply('i', 'j + 3'
...  ).redimension(db.arrays.foo
...  ).insert(db.arrays.foo)

>>> db.arrays.foo[:]
   i    x
0  0  0.0
1  1  1.0
2  2  2.0
3  3  3.0
4  4  4.0
5  5  5.0

>>> db.input('<i:int64 not null, x:int64>[j]',
...          upload_data=db.arrays.foo.fetch(as_dataframe=False)
...  ).redimension(db.arrays.foo
...  ).store('bar')
... # doctest: +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'bar')

>>> numpy.all(db.arrays.bar.fetch(as_dataframe=False)
...           == db.arrays.foo.fetch(as_dataframe=False))
True

>>> buf = numpy.array([bytes([10, 20, 30])], dtype='object')

>>> db.input('<b:binary not null>[i]', upload_data=buf).store('taz')
... # doctest: +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'taz')

>>> db.load('taz',
...         upload_data=buf,
...         upload_schema=Schema.fromstring('<b:binary not null>[i]'))
... # doctest: +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'taz')

For files already available on the server the ``input`` or ``load``
operators can be invoked with the full set of arguments supported by
SciDB. Arguments that need to be quoted in SciDB need to be
double-quoted in SciDB-Py. For example:

>>> db.load('foo', "'/data.csv'", 0, "'CSV'")
... # doctest: +SKIP

>>> for ar in ['foo', 'bar', 'taz']: db.remove(ar)

The ``store`` function accepts a ``temp`` argument as well. If the
``temp`` argument is set to ``True``, a temporary array is created.

>>> ar = db.build('<x:int64>[i=0:2]', 'i').store('foo', temp=True)
>>> db.list()[['name', 'temporary']]
  name  temporary
0  foo       True
>>> db.remove(ar)

If an array name is not specified for the ``store`` operator, an array
name is generated. Arrays with generated names are removed when the
returned Array object is garbage collected. This behavior can be
changed by specifying the ``gc=False`` argument to the store operator.

>>> ar = db.input(upload_data=numpy.arange(3)).store()
>>> ar
... # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
Array(DB('http://localhost:8080',
         None,
         None,
         None,
         False,
         None,
         False,
         256,
         False),
      'py_...')
>>> del ar


The iquery Function
===================

Use the :meth:`DB.iquery()<scidbpy.db.DB.iquery>` function to execute
literal queries against SciDB:

>>> db.iquery('store(build(<x:int64>[i=0:2], i), foo)')


Download Data from SciDB
------------------------

The ``iquery`` function can be used to download data from SciDB by
specifying the ``fetch=True`` argument:

>>> db.iquery('scan(foo)', fetch=True)
   i    x
0  0  0.0
1  1  1.0
2  2  2.0

To avoid downloading the dimension information and only download the
attributes, use the ``atts_only=True`` argument:

>>> db.iquery('scan(foo)', fetch=True, atts_only=True)
     x
0  0.0
1  1.0
2  2.0

>>> db.iquery('remove(foo)')


Download operator output directly:

>>> db.iquery('build(<x:int64 not null>[i=0:2], i)',
...           fetch=True)
   i  x
0  0  0
1  1  1
2  2  2

>>> db.iquery('build(<x:int64 not null>[i=0:2], i)',
...           fetch=True,
...           atts_only=True)
   x
0  0
1  1
2  2


If dimension names collide with attribute names, unique dimension
names are created:

>>> db.iquery('apply(build(<x:int64 not null>[i=0:2], i), i, i)', fetch=True)
   i_1  x  i
0    0  0  0
1    1  1  1
2    2  2  2


If schema is known, it can be provided to ``iquery`` using the
``schema`` argument. This speeds up the execution as ``iquery`` does
not need to issue a ``show()`` query first in order to determine the
schema:

>>> from scidbpy import Schema
>>> iquery(db,
...        'build(<x:int64 not null>[i=0:2], i)',
...        fetch=True,
...        schema=Schema(None,
...                      (Attribute('x', 'int64', not_null=True),),
...                      (Dimension('i', 0, 2),)))
   i  x
0  0  0
1  1  1
2  2  2

>>> iquery(db,
...        'build(<x:int64 not null>[i=0:2], i)',
...        fetch=True,
...        atts_only=True,
...        schema=Schema.fromstring('<x:int64 not null>[i=0:2]'))
... # doctest: +NORMALIZE_WHITESPACE
   x
0  0
1  1
2  2


Attributes with null-able types are promoted as per Pandas `promotion
scheme <http://pandas.pydata.org/pandas-docs/stable/gotchas.html
#na-type-promotions>`_ During such promotions, precision loss might
occur. The user is notified about this via a ``UserWarning``. If
desired, such warnings can be disabled as follows:

>>> import warnings
>>> warnings.filterwarnings('ignore', category = UserWarning)

Type promotions can be avoided altogether, by specifying
``dataframe_promo = False`` in the ``iquery`` arguments. In this case,
object records will be used instead of atomic values:

>>> iquery(db,
...        'build(<x:int64>[i=0:2], i)',
...        fetch=True,
...        atts_only=True,
...        as_dataframe=True,
...        dataframe_promo=False)
... # doctest: +NORMALIZE_WHITESPACE
   x
0  (255, 0)
1  (255, 1)
2  (255, 2)


Download as NumPy Array:

>>> iquery(db,
...        'build(<x:int64>[i=0:2], i)',
...        fetch=True,
...        as_dataframe=False)
... # doctest: +NORMALIZE_WHITESPACE
array([(0, (255, 0)), (1, (255, 1)), (2, (255, 2))],
      dtype=[('i', '<i8'), ('x', [('null', 'u1'), ('val', '<i8')])])

>>> iquery(db,
...        'build(<x:int64>[i=0:2], i)',
...        fetch=True,
...        atts_only=True,
...        as_dataframe=False)
... # doctest: +NORMALIZE_WHITESPACE
array([((255, 0),), ((255, 1),), ((255, 2),)],
      dtype=[('x', [('null', 'u1'), ('val', '<i8')])])


If the `accelerated_io_tools
<https://github.com/Paradigm4/accelerated_io_tools>`_ SciDB plugin is
installed and enabled in `Shim <https://github.com/Paradigm4/shim>`_,
SciDB arrays can be downloaded using the Apache Arrow library:

>>> db.iquery('scan(foo)', fetch=True, use_arrow=True)
... # doctest: +SKIP
   i    x
0  0  0.0
1  1  1.0
2  2  2.0


By default, Apache Arrow is not used. If desired, ``use_arrow`` can be
set to ``True`` at connection time (it is ``False`` by default). Once
set at connection time, this value is used for any subsequent
``iquery`` calls if not overwritten:

>>> db_arrow = connect(use_arrow=True)
>>> db_arrow.iquery('scan(foo)', fetch=True)
... # doctest: +SKIP
   i    x
0  0  0.0
1  1  1.0
2  2  2.0


Upload Data to SciDB
--------------------

Data can be uploaded using the ``iquery`` function by providing an
``upload_data`` argument. A file name placeholder needs to be provided
as part of the SciDB query string. The upload array schema and data
format can be provided explicitly or as placeholders in the query
string. The placeholders are replaced with the explicit values by the
``iquery`` function, before the query is sent to SciDB.

The SciDB query placeholders are:

* ``'{fn}'``: **mandatory** placeholder which is replaced with the
  file name of the server file where the uploaded data is stored. It
  has to be *quoted* with single quotes in the query string.

* ``{sch}``: *optional* placeholder which is replaced with the upload
  array schema. It does *not* need to be quoted.

* ``'{fmt}'``: *optional* placeholder which is replaced with the
  upload array format.  It has to be *quoted* with single quotes in
  the query string.

See examples in the following subsections.


Upload NumPy Arrays
^^^^^^^^^^^^^^^^^^^

Provide a SciDB ``input``, ``store``, ``insert``, or ``load`` query
and a NumPy array. If the schema or format are provided as
placeholders, the upload data *dtype* or upload schema is used to
populate these placeholders.

>>> db.iquery("store(input(<x:int64>[i], '{fn}', 0, '{fmt}'), foo)",
...           upload_data=numpy.arange(3))

>>> db.arrays.foo[:]
   i    x
0  0  0.0
1  1  1.0
2  2  2.0

>>> db.iquery("insert(input({sch}, '{fn}', 0, '(int64)'), foo)",
...           upload_data=numpy.arange(3))

Optionally, a ``Schema`` object can be used to specify the upload
schema using the ``upload_schema`` argument:

>>> db.iquery("load(foo, '{fn}', 0, '{fmt}')",
...           upload_data=numpy.arange(3),
...           upload_schema=Schema.fromstring('<x:int64 not null>[i]'))


Upload Binary Data
^^^^^^^^^^^^^^^^^^

Provide a SciDB ``input``, ``store``, ``insert``, or ``load`` query
and binary data. The schema of the upload data needs to be provided
either explicitly in the query string or using the ``upload_schema``
argument. If the schema is not provides using the ``upload_schema``
argument, the format needs to be provided explicitly in the query
string:

>>> db.iquery("store(input({sch}, '{fn}', 0, '{fmt}'), foo)",
...           upload_data=numpy.arange(3).tobytes(),
...           upload_schema=Schema.fromstring('<x:int64 not null>[i]'))

>>> db.iquery("insert(input(foo, '{fn}', 0, '(int64)'), foo)",
...           upload_data=numpy.arange(3).tobytes())

>>> db.iquery("load(foo, '{fn}', 0, '(int64)')",
...           upload_data=numpy.arange(3).tobytes())


Upload Data Files
^^^^^^^^^^^^^^^^^

A binary or text file-like object can be used to specify the upload
data. The content of the file has to be in one of the `supported SciDB
formats
<https://paradigm4.atlassian.net/wiki/spaces/ESD169/pages/50856232/input>`_. A
matching format specification has to be provided as well:

>>> with open('array.bin', 'wb') as file:
...     n = file.write(numpy.arange(3).tobytes())

>>> db.iquery("load(foo, '{fn}', 0, '(int64)')",
...           upload_data=open('array.bin', 'rb'))

>>> with open('array.csv', 'w') as file:
...     n = file.write('1\\n2\\n3\\n')

>>> db.iquery("load(foo, '{fn}', 0, 'CSV')",
...           upload_data=open('array.csv', 'r'))

>>> import os
>>> os.remove('array.bin')
>>> os.remove('array.csv')
>>> db.remove(db.arrays.foo)

Please note that the data file is not read into the SciDB-Py
library. The data file object is passed directly to the ``requests``
library which handles the HTTP communication with *Shim*.


Joins
=====

SciDB provides a number of join operators, including ``join``,
``cross_join``, etc. As with any other operators, these operators can
be used as literals in the ``DB.iquery`` function or as functions of
the ``DB`` object. Here are some join examples exposing various
features of the library:

>>> foo = db.build('<val:double>[i=0:2; j=0:2]', 'i * 3 + j').store('foo')

>>> db.join(foo,
...         db.build('<val:double>[i=0:2; j=0:5]', '0'))[:]
   i  j  val  val_1
0  0  0  0.0    0.0
1  0  1  1.0    0.0
2  0  2  2.0    0.0
3  1  0  3.0    0.0
4  1  1  4.0    0.0
5  1  2  5.0    0.0
6  2  0  6.0    0.0
7  2  1  7.0    0.0
8  2  2  8.0    0.0

>>> db.build('<val:double>[i=0:2; j=0:5]', '0'
...  ).join(db.arrays.foo)[:]
   i  j  val  val_1
0  0  0  0.0    0.0
1  0  1  0.0    1.0
2  0  2  0.0    2.0
3  1  0  0.0    3.0
4  1  1  0.0    4.0
5  1  2  0.0    5.0
6  2  0  0.0    6.0
7  2  1  0.0    7.0
8  2  2  0.0    8.0

>>> db.cross_join(foo,
...               db.build('<val:double>[k=0:5]', 'k + 100'),
...               foo.j, 'k')[:]
   i  j  val  val_1
0  0  0  0.0  100.0
1  0  1  1.0  101.0
2  0  2  2.0  102.0
3  1  0  3.0  100.0
4  1  1  4.0  101.0
5  1  2  5.0  102.0
6  2  0  6.0  100.0
7  2  1  7.0  101.0
8  2  2  8.0  102.0

>>> bar = db.build('<val:double>[j=0:5]', 'j + 100').store()

>>> db.cross_join(db.arrays.foo % 'left',
...               bar % 'right',
...               'left.j', 'right.j')[:]
   i  j  val  val_1
0  0  0  0.0  100.0
1  0  1  1.0  101.0
2  0  2  2.0  102.0
3  1  0  3.0  100.0
4  1  1  4.0  101.0
5  1  2  5.0  102.0
6  2  0  6.0  100.0
7  2  1  7.0  101.0
8  2  2  8.0  102.0

>>> db.remove(foo)
>>> db.remove(bar)


SciDB Enterprise Edition Features
=================================

SciDB Enterprise Edition features can be used directly as any other
operators. One special case is the ``set_namespace`` operator. The
operator alters the ``DB`` object on which it is called. The effect
persists until the next call to ``set_namespace``. The operator can be
called directly (i.g., ``db.set_namespace``) or through the ``iquery``
function. No immediate query is executed in SciDB, but the new
namespace will take effect for any subsequent SciDB queries:

>>> print(db)
scidb_url         = http://localhost:8080
scidb_auth        = None
http_auth         = None
verify            = None
admin             = False
namespace         = None
use_arrow         = False
result_size_limit = 256
no_ops            = False

Notice the ``namespace`` field of the ``DB`` instance.

>>> db.set_namespace('private')
... # doctest: +SKIP
>>> print(db)
... # doctest: +SKIP
scidb_url         = http://localhost:8080
scidb_auth        = None
http_auth         = None
verify            = None
admin             = False
namespace         = private
use_arrow         = False
result_size_limit = 256
no_ops            = False
>>> db.show_namespace()[0]['name']['val']
... # doctest: +SKIP
'private'

>>> db.iquery("set_namespace('public')")
>>> print(db)
scidb_url         = http://localhost:8080
scidb_auth        = None
http_auth         = None
verify            = None
admin             = False
namespace         = public
use_arrow         = False
result_size_limit = 256
no_ops            = False
>>> db.show_namespace()[0]['name']['val']
... # doctest: +SKIP
'public'

For convenience, an initial namespace can be provided at connection
time:

>>> db_ssl = connect('https://localhost:8083',
...                  verify=False,
...                  namespace='public')
... # doctest: +SKIP
>>> print(db_ssl)
... # doctest: +SKIP
scidb_url         = https://localhost:8083
scidb_auth        = None
http_auth         = None
verify            = False
admin             = Flase
namespace         = public
use_arrow         = False
result_size_limit = 256
no_ops            = False

"""

from .db import connect, iquery, Array
from .schema import Attribute, Dimension, Schema

__version__ = '19.3.3'



# List of hungry operators, they execute immediately, e.g.,
# ``db.remove``.
#
# Operators **not** listed here are lazy. They will execute at a later
# time.
#
ops_hungry = (
    # list('operators');
    # ---
    'cancel',

    'consume',
    'create_array',
    'create_array_using',

    'delete',

    'help',

    'insert',

    'load_library',
    'load_module',

    'remove',
    'remove_versions',
    'rename',

    'save',

    'store',

    'unload_library',

    # list('macros');
    # ---
    'load',

    # load_library('dense_linear_algebra');
    # ---
    'mpi_init',

    # load_library('namespaces');
    # ---
    'add_user_to_role',
    'change_user',
    'create_namespace',
    'create_role',
    'create_user',
    'drop_namespace',
    'drop_role',
    'drop_user',
    'drop_user_from_role',
    'move_array_to_namespace',
    'set_namespace',
    'set_role_permissions',
    )

# List of operators with string arguments. The list is groupped by
# argument position. If an operator has multiple string arguments, it
# is listed multiple times. If an argument can be a string or somting
# else, it is excluded from this list.
#
string_args = (
    # 1st argument
    set((
        # list('operators');
        # ---
        'cancel',
        'list',
        'load_library',
        'load_module',
        'unload_library',

        # load_library('namespaces');
        # ---
        'add_user_to_role',
        'change_user',
        'create_namespace',
        'create_role',
        'create_user',
        'drop_namespace',
        'drop_role',
        'drop_user',
        'drop_user_from_role',
        'set_namespace',
        'set_role',
        'set_role_permissions',
        'show_users_in_role',
    )),
    # 2nd argument
    set((
        # list('operators');
        # ---
        'input',
        'load',
        'save',
        'show',

        # load_library('namespaces');
        # ---
        'add_user_to_role',
        'change_user',
        'create_user',
        'drop_user_from_role',
        'move_array_to_namespace',
        'set_role_permissions',
    )),
    # 3rd argument
    set((
        # load_library('namespaces');
        # ---
        'change_user',
        'set_role_permissions',
    )),
    # 4thd argument
    set((
        # list('operators');
        # ---
        'input',
        'load',
        'save',

        # load_library('namespaces');
        # ---
        'set_role_permissions',
    )),
    # 5th argument
    set((
    )),
    # 6th argument
    set((
    )),
    # 7th argument
    set((
    )),
    # 8th argument
    set((
    ))
)



"""DB, Array, and Operator
=======================

Classes for connecting to SciDB and executing queries.

"""

import copy
import enum
import itertools
import logging
import numpy
import os
import pandas
import pyarrow
import re
import requests
import string
import threading
import uuid
import warnings

try:
    from weakref import finalize
except ImportError:
    from backports.weakref import finalize

from .meta import ops_hungry, string_args
from .schema import Attribute, Dimension, Schema


class Shim(enum.Enum):
    cancel = 'cancel'
    execute_query = 'execute_query'
    new_session = 'new_session'
    read_bytes = 'read_bytes'
    read_lines = 'read_lines'
    release_session = 'release_session'
    upload = 'upload'


class Password_Placeholder(object):
    def __repr__(self):
        return 'PASSWORD_PROVIDED'


def _shim_release_session(scidb_url, http_auth, verify, id):
    """Make Shim release_session request"""
    url = requests.compat.urljoin(scidb_url, Shim.release_session.value)
    req = requests.get(
        url,
        params={'id': id},
        auth=http_auth,
        verify=verify)
    req.reason = req.content
    req.raise_for_status()


class DB(object):
    """SciDB Shim connection object.

    >>> DB()
    ... # doctest: +NORMALIZE_WHITESPACE
    DB('http://localhost:8080',
       None,
       None,
       None,
       False,
       None,
       False,
       256,
       False)

    >>> print(DB())
    scidb_url         = http://localhost:8080
    scidb_auth        = None
    http_auth         = None
    verify            = None
    admin             = False
    namespace         = None
    use_arrow         = False
    result_size_limit = 256
    no_ops            = False

    Constructor parameters:

    :param string scidb_url: SciDB connection URL. The URL for the
      Shim server. If ``None``, use the value of the ``SCIDB_URL``
      environment variable, if present (default
      ``http://localhost:8080``)

    :param tuple scidb_auth: Tuple with username and password for
      connecting to SciDB, if password authentication method is used
      (default ``None``)

    :param tuple http_auth: Tuple with username and password for
      connecting to Shim, if Shim authentication is used (default
      ``None``)

    :param bool verify: If ``False``, HTTPS certificates are not
      verified. This value is passed to the Python ``requests``
      library. See Python `requests
      <http://docs.python-requests.org/en/master/>`_ library `SSL Cert
      Verification
      <http://docs.python-requests.org/en/master/user/advanced/
      #ssl-cert-verification>`_ section for details on the ``verify``
      argument (default ``None``)

    :param bool admin: Set to ``True`` to open a higher-priority
      session. This is identical with the ``--admin`` flag for the
      ``iquery`` SciDB client, see `SciDB Documentation
      <https://paradigm4.atlassian.net/wiki/spaces/scidb>`_ for
      details (default ``False``)

    :param string namespace: Initial namespace for the
      connection. Only applicable for SciDB Enterprise Edition. The
      namespace can changed at any time using the ``set_namespace``
      SciDB operator (default ``None``)

    :param bool use_arrow: If ``True``, download SciDB array using
      Apache Arrow library. Requires ``accelerated_io_tools`` and
      ``aio`` enabled in ``Shim``. If ``True``, a Pandas DataFrame is
      returned (``as_dataframe`` has no effect) and null-able types
      are promoted as per Pandas `promotion scheme
      <http://pandas.pydata.org/pandas-docs/stable/gotchas.html
      #na-type-promotions>`_ (``dataframe_promo`` has no effect). It
      can be overriden for each ``iquery`` call (default ``False``)

    :param int result_size_limit: absolute limit of the output file in
      Megabytes. Effective only when the `accelerated_io_tools`
      plug-in is installed in SciDB and `aio` is enabled in Shim
      (default `256` MB)

    :param bool no_ops: If ``True``, the list of operators is not
      fetched at this time and the connection is not implicitly
      verified. This expedites the execution of the function but
      disallows for calling the SciDB operators directly from the
      ``DB`` instance e.g., ``db.scan`` (default ``False``)

    """

    _show_query = "show('{}', 'afl')"

    def __init__(
            self,
            scidb_url=None,
            scidb_auth=None,
            http_auth=None,
            verify=None,
            admin=False,
            namespace=None,
            use_arrow=False,
            result_size_limit=256,
            no_ops=False):
        if scidb_url is None:
            scidb_url = os.getenv('SCIDB_URL', 'http://localhost:8080')

        self.scidb_url = scidb_url
        self.verify = verify
        self.admin = admin
        self.namespace = namespace
        self.use_arrow = use_arrow
        self.result_size_limit = result_size_limit
        self.no_ops = no_ops

        if http_auth:
            self._http_auth = requests.auth.HTTPDigestAuth(*http_auth)
            self.http_auth = (http_auth[0], Password_Placeholder())
        else:
            self._http_auth = self.http_auth = None

        admin_shim = 1 if self.admin else 0
        if scidb_auth:
            self._id = self._shim(Shim.new_session,
                                  user=scidb_auth[0],
                                  password=scidb_auth[1],
                                  admin=admin_shim).text
            self.scidb_auth = (scidb_auth[0], Password_Placeholder())
        else:
            self._id = self._shim(Shim.new_session,
                                  admin=admin_shim).text
            self.scidb_auth = None

        finalize(self,
                 _shim_release_session,
                 self.scidb_url,
                 self._http_auth,
                 self.verify,
                 self._id)

        self.arrays = Arrays(self)

        self._uid = uuid.uuid1().hex
        self._lock = threading.Lock()
        self._array_cnt = 0
        self._formatter = string.Formatter()

        if self.no_ops:
            self.operators = None
            self._dir = None
        else:
            self.load_ops()

    def __iter__(self):
        return (i for i in (
            self.scidb_url,
            self.scidb_auth,
            self.http_auth,
            self.verify,
            self.admin,
            self.namespace,
            self.use_arrow,
            self.result_size_limit,
            self.no_ops))

    def __repr__(self):
        return ('{}({!r}, ' +
                '{!r}, ' +
                '{!r}, ' +
                '{!r}, ' +
                '{!r}, ' +
                '{!r}, ' +
                '{!r}, ' +
                '{!r}, ' +
                '{!r})').format(
            type(self).__name__, *self)

    def __str__(self):
        return '''\
scidb_url         = {}
scidb_auth        = {}
http_auth         = {}
verify            = {}
admin             = {}
namespace         = {}
use_arrow         = {}
result_size_limit = {}
no_ops            = {}'''.format(*self)

    def __getattr__(self, name):
        if self.operators and name in self.operators:
            return Operator(self, name)
        elif self.no_ops:
            raise AttributeError(
                ("Operators not loaded. Run 'load_ops()' or " +
                 "use 'no_ops=False' (default) " +
                 "at connection time (constructor)").format(
                     type(self)))
        else:
            raise AttributeError(
                '{.__name__!r} object has no attribute {!r}'.format(
                    type(self), name))

    def __dir__(self):
        return self._dir

    def iquery(self,
               query,
               fetch=False,
               use_arrow=None,
               atts_only=False,
               as_dataframe=True,
               dataframe_promo=True,
               schema=None,
               upload_data=None,
               upload_schema=None):
        """Execute query in SciDB

        :param string query: SciDB AFL query to execute

        :param bool fetch: If ``True``, download SciDB array (default
          ``False``)

        :param bool use_arrow: If ``True``, download SciDB array using
          Apache Arrow library. Requires ``accelerated_io_tools`` and
          ``aio`` enabled in ``Shim``. If ``True``, a Pandas DataFrame
          is returned (``as_dataframe`` has no effect) and null-able
          types are promoted as per Pandas `promotion scheme
          <http://pandas.pydata.org/pandas-docs/stable/gotchas.html
          #na-type-promotions>`_ (``dataframe_promo`` has no
          effect). If ``None`` the ``use_arrow`` value set at
          connection time is used (default ``None``)

        :param bool atts_only: If ``True``, download only SciDB array
          attributes without dimensions (default ``False``)

        :param bool as_dataframe: If ``True``, return a Pandas
          DataFrame. If ``False``, return a NumPy array (default
          ``True``)

        :param bool dataframe_promo: If ``True``, null-able types are
          promoted as per Pandas `promotion scheme
          <http://pandas.pydata.org/pandas-docs/stable/gotchas.html
          #na-type-promotions>`_ If ``False``, object records are used
          for null-able types (default ``True``)

        :param schema: Schema of the SciDB array to use when
          downloading the array. Schema is not verified. If schema is
          a Schema instance, it is copied. Otherwise, a
          :py:class:``Schema`` object is built using
          :py:func:``Schema.fromstring`` (default ``None``)

        >>> DB().iquery('build(<x:int64>[i=0:1; j=0:1], i + j)', fetch=True)
           i  j    x
        0  0  0  0.0
        1  0  1  1.0
        2  1  0  1.0
        3  1  1  2.0

        >>> DB().iquery("input({sch}, '{fn}', 0, '{fmt}')",
        ...             fetch=True,
        ...             upload_data=numpy.arange(3, 6))
           i  x
        0  0  3
        1  1  4
        2  2  5

        """
        # Set use_arrow using local/global
        if use_arrow is None:
            use_arrow = self.use_arrow

        # Special case: -- - set_namespace - --
        if query.startswith('set_namespace(') and query[-1] == ')':
            param = query[len('set_namespace('):-1]
            # Unquote if quoted. Will be quoted when set in prefix.
            if param[0] == "'" and param[-1] == "'":
                param = param[1:-1]
            self.namespace = param
            return

        if upload_data is not None:
            if isinstance(upload_data, numpy.ndarray):
                if upload_schema is None:
                    try:
                        upload_schema = Schema.fromdtype(upload_data.dtype)
                    except Exception as e:
                        warnings.warn(
                            'Mapping NumPy dtype to SciDB schema failed. ' +
                            'Try providing an explicit upload_schema')
                        raise e

                # Convert upload data to bytes
                if upload_schema.is_fixsize():
                    upload_data = upload_data.tobytes()
                else:
                    upload_data = upload_schema.tobytes(upload_data)

            # Check if placeholders are present
            place_holders = set(
                field_name
                for _1, field_name, _3, _4 in self._formatter.parse(query))
            if 'fn' not in place_holders:
                warnings.warn(
                    'upload_data provided, but {fn} placeholder is missing',
                    stacklevel=2)
            if 'fmt' in place_holders and upload_schema is None:
                warnings.warn(
                    'upload_data and {fmt} placeholder provided, ' +
                    'but upload_schema is None',
                    stacklevel=2)

            # Check if upload data is bytes or file-like object
            if not (isinstance(upload_data, bytes) or
                    isinstance(upload_data, bytearray) or
                    hasattr(upload_data, 'read')):
                warnings.warn(
                    'upload_data is not bytes or file-like object',
                    stacklevel=2)

            fn = self._shim(Shim.upload, data=upload_data).text
            query = query.format(
                sch=upload_schema,
                fn=fn,
                fmt=upload_schema.atts_fmt_scidb if upload_schema else None)

        if fetch:
            # Use provided schema or get schema from SciDB
            if schema:
                # Deep-copy schema since we might be mutating it
                if isinstance(schema, Schema):
                    if not atts_only and not use_arrow:
                        schema = copy.deepcopy(schema)
                else:
                    schema = Schema.fromstring(schema)
            else:
                # Execute 'show(...)' and Download text
                self._shim(
                    Shim.execute_query,
                    query=DB._show_query.format(query.replace("'", "\\'")),
                    save='tsv')
                schema = Schema.fromstring(
                    self._shim(Shim.read_lines, n=0).text)

            # Attributes and dimensions can collide. Run make_unique to
            # remove any collisions.
            #
            # make_unique fixes any collision, but if we don't
            # download the dimensions, we don't need to fix collisions
            # between dimensions and attributes. So, we use
            # make_unique only if there are collisions within the
            # attribute names.
            if ((not atts_only or
                 len(set((a.name for a in schema.atts))) <
                 len(schema.atts)) and schema.make_unique()):
                # Dimensions or attributes were renamed due to
                # collisions. We need to cast.
                query = 'cast({}, {:h})'.format(query, schema)

            # Unpack
            if not atts_only and not use_arrow:
                # apply: add dimensions as attributes
                # project: place dimensions first
                query = 'project(apply({}, {}), {})'.format(
                    query,
                    ', '.join('{0}, {0}'.format(d.name) for d in schema.dims),
                    ', '.join(i.name for i in itertools.chain(
                        schema.dims, schema.atts)))

                # update schema after apply
                schema.make_dims_atts()

            # Execute Query and Download content
            self._shim(Shim.execute_query,
                       query=query,
                       save='arrow' if use_arrow else schema.atts_fmt_scidb,
                       result_size_limit=self.result_size_limit,
                       atts_only=1 if atts_only or not use_arrow else 0)
            buf = self._shim(Shim.read_bytes, n=0).content

            # Build result
            if use_arrow:
                data = pyarrow.RecordBatchStreamReader(
                    pyarrow.BufferReader(buf)).read_pandas()
                # Place dimensions first
                if not atts_only:
                    data = data[[i.name for i in itertools.chain(
                        schema.dims, schema.atts)]]

            elif schema.is_fixsize():
                data = numpy.frombuffer(buf, dtype=schema.atts_dtype)

                if as_dataframe:
                    data = pandas.DataFrame.from_records(data)

                    if dataframe_promo:
                        schema.promote(data)
            else:
                # Parse binary buffer
                data = schema.frombytes(buf, as_dataframe, dataframe_promo)

                if as_dataframe:
                    data = pandas.DataFrame.from_records(data)

            return data

        else:                   # fetch=False
            self._shim(Shim.execute_query, query=query)

            # Special case: -- - load_library - --
            if query.startswith('load_library('):
                self.load_ops()

    def iquery_readlines(self, query):
        """Execute query in SciDB

        >>> DB().iquery_readlines('build(<x:int64>[i=0:2], i * i)')
        ... # doctest: +ELLIPSIS
        [...'0', ...'1', ...'4']

        >>> DB().iquery_readlines(
        ...   'apply(build(<x:int64>[i=0:2], i), y, i + 10)')
        ... # doctest: +ELLIPSIS
        [[...'0', ...'10'], [...'1', ...'11'], [...'2', ...'12']]
        """
        self._shim(Shim.execute_query, query=query, save='tsv')
        ret = self._shim_readlines()
        return ret

    def next_array_name(self):
        """Generate a uniqu array name. Keep track on these names using the
           _uid field and a counter
        """
        # Thread-safe counter
        with self._lock:
            self._array_cnt += 1
            return 'py_{}_{}'.format(self._uid, self._array_cnt)

    def load_ops(self):
        """Get list of operators and macros.
        """
        query_id = self._shim(
            Shim.execute_query,
            query="project(list('operators'), name)",
            save='tsv').text  # set query ID as DB instance ID
        operators = self._shim_readlines()

        self._shim(
            Shim.execute_query,
            query="project(list('macros'), name)",
            save='tsv').content
        macros = self._shim_readlines()

        self.no_ops = False
        self.operators = operators + macros
        self._dir = (self.operators +
                     ['arrays',
                      'gc',
                      'iquery',
                      'iquery_readlines',
                      'upload'])
        self._dir.sort()

    def _cleanup(self, endpoint):
        """Attempt to cleanup in case of user interruption or error. Send a
           cancel request to the query associated with the current
           session

        """
        if endpoint == Shim.execute_query:
            requests.get(
                requests.compat.urljoin(self.scidb_url, Shim.cancel.value),
                params={'id': self._id},
                auth=self._http_auth,
                verify=self.verify)

    def _shim(self, endpoint, **kwargs):
        """Make request on Shim endpoint"""

        if endpoint != Shim.new_session:
            kwargs.update(id=self._id)

        if self.use_arrow and endpoint == Shim.execute_query:
            kwargs['result_size_limit'] = self.result_size_limit

        # Add prefix to request, if necessary
        if self.namespace and endpoint == Shim.execute_query:
            kwargs['prefix'] = "set_namespace('{}')".format(self.namespace)

        # Make request
        url = requests.compat.urljoin(self.scidb_url, endpoint.value)
        if endpoint == Shim.upload:  # Post request
            req = requests.post(
                '{}?id={}'.format(url, kwargs['id']),
                data=kwargs['data'],
                auth=self._http_auth,
                verify=self.verify)
        else:                        # Get request
            try:
                req = requests.get(
                    url,
                    params=kwargs,
                    auth=self._http_auth,
                    verify=self.verify)
            except KeyboardInterrupt as e:
                self._cleanup(endpoint)
                raise e

        req.reason = req.content
        try:
            req.raise_for_status()
        except Exception as e:
            self._cleanup(endpoint)
            raise e
        return req

    def _shim_readlines(self):
        """Read data from Shim and parse as text lines"""
        return [line.split('\t') if '\t' in line else line
                for line in self._shim(
                        Shim.read_lines, n=0).text.splitlines()]


class Arrays(object):
    """Access to arrays available in SciDB"""
    def __init__(self, db):
        self.db = db

    def __repr__(self):
        return '{}({!r})'.format(
            type(self).__name__, self.db)

    def __str__(self):
        return '''DB:
{}'''.format(self.db)

    def __getattr__(self, name):
        """db.arrays.foo"""
        return Array(self.db, name)

    def __getitem__(self, name):
        """db.arrays['foo']"""
        return Array(self.db, name)

    def __dir__(self):
        """Download the list of SciDB arrays. Use 'project(list(), name)' to
        download only names and schemas
        """
        return self.db.iquery_readlines('project(list(), name)')


class Array(object):
    """Access to individual array"""
    def __init__(self, db, name, gc=False):
        self.db = db
        self.name = name

        if gc:
            finalize(self,
                     self.db.iquery,
                     'remove({})'.format(self.name))

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            type(self).__name__, self.db, self.name)

    def __str__(self):
        return self.name

    def __getattr__(self, key):
        return ArrayExp('{}.{}'.format(self.name, key))

    def __getitem__(self, key):
        return self.fetch()[key]

    def __dir__(self):
        """Download the schema of the SciDB array, using ``show()``"""
        sh = Schema.fromstring(
            self.db.iquery_readlines('show({})'.format(self))[0][0])
        ls = [i.name for i in itertools.chain(sh.atts, sh.dims)]
        ls.sort()
        return ls

    def __mod__(self, alias):
        """Overloads ``%`` operator to add support for aliasing"""
        return Array(self.db, '{} as {}'.format(self.name, alias))

    def fetch(self, **kwargs):
        return self.db.iquery('scan({})'.format(self),
                              fetch=True,
                              **kwargs)

    def head(self, n=5, **kwargs):
        """Similar to ``pandas.DataFrame.head``. Makes use of the ``limit``
        operator, if available.

        """
        if 'limit' in self.db.operators:
            return self.db.iquery('limit({}, {})'.format(self, n),
                                  fetch=True,
                                  **kwargs)
        else:
            warnings.warn(
                '"limit" operator not available. ' +
                'Fetching the entire array. ' +
                'See https://github.com/Paradigm4/limit.')
            return self.fetch(**kwargs)[:n]

    def schema(self):
        return Schema.fromstring(
            self.db.iquery_readlines("show({})".format(self))[0][0])


class ArrayExp(object):
    """Access to individual attribute or dimension"""
    def __init__(self, exp):
        self.exp = exp

    def __repr__(self):
        return '{}({!r})'.format(type(self).__name__, self.exp)

    def __str__(self):
        return '{}'.format(self.exp)

    def __add__(self, other):
        return ArrayExp('{} + {}'.format(self, other))


class Operator(object):
    """Store SciDB operator and arguments. Hungry operators (e.g., remove,
    store, etc.) evaluate immediately. Lazy operators evaluate on data
    fetch.

    """
    def __init__(self, db, name, upload_data=None, upload_schema=None, *args):
        self.db = db
        self.name = name.lower()
        self.upload_data = upload_data
        self.upload_schema = upload_schema

        self.args = list(args)
        self.is_lazy = self.name not in ops_hungry

        self._dir = self.db.operators + ['fetch']
        self._dir.sort()

    def __repr__(self):
        return '{}(db={!r}, name={!r}, args=[{}])'.format(
            type(self).__name__,
            self.db,
            self.name,
            ', '.join('{!r}'.format(i) for i in self.args))

    def __str__(self):
        args_fmt = []
        for (pos, arg) in enumerate(self.args):
            # Format argument to string (possibly recursive)
            arg_fmt = '{}'.format(arg)

            # Special case: quote string argument if not quoted
            if (pos < len(string_args) and
                    self.name in string_args[pos] and
                    len(arg) and
                    arg_fmt[0] != "'" and
                    arg_fmt[-1] != "'"):

                arg_fmt = "'{}'".format(arg_fmt)

            # Special case: -- - show - --
            if (pos == 0 and
                    self.name == 'show' and
                    (isinstance(arg, Operator) or
                     len(self.args) > 1)):
                arg_fmt = "'{}'".format(arg_fmt.replace("'", "\\'"))

            # Add to arguments list
            args_fmt.append(arg_fmt)

        return '{}({})'.format(self.name, ', '.join(args_fmt))

    def __call__(self, *args, **kwargs):
        """Returns self for lazy expressions. Executes immediate expressions.
        """
        # Propagate "upload_data" and "upload_schema" from previous
        # operators
        for arg in args:
            if isinstance(arg, Operator) and (
                    arg.upload_data is not None or
                    arg.upload_schema is not None):
                if self.upload_data is not None or (self.upload_schema
                                                    is not None):
                    raise NotImplementedError(
                        'Queries with multiple "upload_data" or ' +
                        '"upload_schema" arguments are not supported')
                self.upload_data = arg.upload_data
                self.upload_schema = arg.upload_schema

        self.args.extend(args)

        # Special case: -- - create_array - --
        if self.name == 'create_array' and len(self.args) < 3:
            # Set "temporary"
            self.args.append(False)

        # Special case: -- - input & load - --
        elif self.name in ('input', 'load'):
            ln = len(self.args)

            # Set upload data
            if 'upload_data' in kwargs.keys():
                self.upload_data = kwargs['upload_data']
            # Set upload schema
            if 'upload_schema' in kwargs.keys():
                # Pass through if provided as argument
                self.upload_schema = kwargs['upload_schema']
            if self.upload_schema is None:
                # If the upload_data is a NumPy array try to map the
                # array dtype to upload schema
                if (self.upload_data is not None and
                        isinstance(self.upload_data, numpy.ndarray)):
                    try:
                        self.upload_schema = Schema.fromdtype(
                            self.upload_data.dtype)
                    except Exception:
                        # Might fail if the dtype contains
                        # objects. The same type mapping is attempted
                        # later in iquery, but there the exception is
                        # propagated
                        pass
                # If the oprator is input, try to map first argument
                # to upload schema
                if (self.upload_schema is None and
                        self.name == 'input' and
                        ln >= 1):
                    try:
                        self.upload_schema = Schema.fromstring(args[0])
                    except Exception:
                        # Fails if the argument is an array name
                        pass

            # Set required arguments if missing
            # Check if "input_file" is present (2nd argument)
            if ln < 2:
                # Check if "existing_array|anonymous_schema"
                # is present (1st argument)
                if ln < 1:
                    self.args.append('{sch}')  # anonymous_schema
                self.args.append("'{fn}'")     # input_file

            # Set optional arguments if missing and necessary
            # Check if "format" is present (4th argument)
            if ln < 4 and self.upload_data is not None:
                # Check if "instance_id" is present (3nd argument)
                if ln < 3:
                    self.args.append(0)        # instance_id
                self.args.append("'{fmt}'")    # format

        # Special case: -- - store - --
        elif self.name == 'store':
            if len(self.args) < 2:
                # Set "named_array"
                self.args.append(self.db.next_array_name())
                # Garbage collect (if not specified)
                if 'gc' not in kwargs.keys():
                    kwargs['gc'] = True
            # If temp=True in kwargs, create a temporary array first
            if 'temp' in kwargs.keys() and kwargs['temp'] is True:
                # Get the schema of the new array
                try:
                    new_schema = Schema.fromstring(
                        self.db.iquery_readlines(
                            "show('{}', 'afl')".format(
                                str(self.args[0]).replace("'", "\\'")))[0][0])
                except requests.HTTPError as e:
                    e.args = (
                        '"temp=True" not supported for complex queries\n' +
                        e.args[0],
                    )
                    raise
                # Set array name
                new_schema.name = self.args[1]
                # Create temporary array
                self.db.iquery('create temp array {}'.format(new_schema))

        # Lazy or hungry
        if self.is_lazy:        # Lazy
            return self

        else:                   # Hungry
            # Execute query
            self.db.iquery(str(self),
                           upload_data=self.upload_data,
                           upload_schema=self.upload_schema)

            # Handle output
            # Special case: -- - load - --
            if self.name == 'load':
                if isinstance(self.args[0], Array):
                    return self.args[0]
                else:
                    return Array(self.db, self.args[0])

            # Special case: -- - store - --
            elif self.name == 'store':
                if isinstance(self.args[1], Array):
                    return self.args[1]
                else:
                    return Array(self.db,
                                 self.args[1],
                                 kwargs.get('gc', False))

    def __getitem__(self, key):
        return self.fetch()[key]

    def __getattr__(self, name):
        if name in self.db.operators:
            return Operator(
                self.db, name, self.upload_data, self.upload_schema, self)
        else:
            raise AttributeError(
                '{.__name__!r} object has no attribute {!r}'.format(
                    type(self), name))

    def __dir__(self):
        return self._dir

    def __mod__(self, alias):
        """Overloads ``%`` operator to add support for aliasing"""
        return Array(self.db, '{} as {}'.format(self, alias))

    def fetch(self, **kwargs):
        if self.is_lazy:
            return self.db.iquery(str(self),
                                  fetch=True,
                                  upload_data=self.upload_data,
                                  upload_schema=self.upload_schema,
                                  **kwargs)

    def schema(self):
        if self.is_lazy:
            return Schema.fromstring(
                self.db.iquery_readlines(
                    "show('{}', 'afl')".format(self))[0][0])


connect = DB
iquery = DB.iquery


if __name__ == "__main__":
    # logging.basicConfig(level=logging.DEBUG)
    import doctest
    doctest.testmod(optionflags=doctest.REPORT_ONLY_FIRST_FAILURE)
